/*************************************************
opengate server
  module for communication with java 

Copyright (C) 1999 Opengate Project Team
Written by Yoshiaki Watanabe
Modified Katsuhiko Eguchi, 2005 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/

#include	"opengatesrv.h"

static void intervalAlarm(int signo);
static void timeoutAlarm(int signo);
static void delayAlarm(int signo);
void getPeerAddr(int sockfd, char *peerAddr);
void GetPeerAddr(int sockfd, char *peerAddr);
void sendTerminateReply(void);
void SendTerminateReply(void);

extern char ruleNumber4[WORDMAXLN];  /* ipfw rule number in string form  */
extern char ruleNumber6[WORDMAXLN];  /* ip6fw rule number in string form */

int listenfd[2]; /* file descriptor for listen port */
int connfd;   /* file descriptor for connection port */
int intervalAlarmRinged=0; /* interval alarm ringed */
int timeoutAlarmRinged=0;  /* timeout alarm ringed */
int delayAlarmRinged=0;    /* delay alarm ringed */

/***************************************/
/* get temp listen port of this server */
/***************************************/
int getListenPort(void)
{
  struct sockaddr_in servaddr4;
  struct sockaddr_in6 servaddr6;
  extern const struct in6_addr in6addr_any;

  int portNo;
  int portmin;
  int portmax;
  
  bzero(&servaddr4, sizeof(servaddr4));
  bzero(&servaddr6, sizeof(servaddr6));

  servaddr4.sin_family=AF_INET;
  servaddr4.sin_addr.s_addr=htonl(INADDR_ANY);
  
  servaddr6.sin6_family=AF_INET6;
  servaddr6.sin6_addr=in6addr_any;

  /* get port range from config file */
  portmin=atoi(GetConfValue("ListenPort/Min"));
  portmax=atoi(GetConfValue("ListenPort/Max"));
  
  /* search unused port between PORTMIN and PORTMAX */
  for(portNo=portmin; portNo<=portmax; portNo++){
    servaddr4.sin_port=htons(portNo);
    servaddr6.sin6_port=htons(portNo);

    listenfd[0]=Socket(AF_INET, SOCK_STREAM, 0);
    listenfd[1]=Socket(AF_INET6, SOCK_STREAM, 0);

    if(listenfd[0]<0 || listenfd[1]<0){ /* if error, return */
      return -1;
    }

    if(listenfd[0]>=FD_SETSIZE && listenfd[1]>=FD_SETSIZE){
      return -1;
    }

    if((bind(listenfd[0], (SA *)&servaddr4, sizeof(servaddr4))==0) &&
       (bind(listenfd[1], (SA *)&servaddr6, sizeof(servaddr6))==0) ){
      break;
    }
       
    Close(listenfd[0]);
    Close(listenfd[1]);
  }

  if(portNo>portmax) return -1;  /* cannot get unused port */
  
  if(Listen(listenfd[0], LISTENQ)<0) return -1; /* if error, return */
  if(Listen(listenfd[1], LISTENQ)<0) return -1;
  
  return portNo;
}

/************************************/
/* wait applet connection           */
/************************************/
int waitAppletConnect(char *userid, char *clientAddr4, char *clientAddr6, int duration, char *macAddr4, int ipStatus, struct clientAddr *pClientAddr)
{
  socklen_t len;
  fd_set rfd, rfd0;                 /* fd_set for select */
  int n,i;                          /* counter */
  int smax;                         /* select max descliptor */
  int ipType;                         /* using IP type */
  struct sockaddr_storage cliaddr;  /* client IP adddress */
  char buff[BUFFMAXLN];             /* read in buffer */
  Sigfunc *sigfunc;                 /* signal function save */
  int connect=0;                    /* java cconnect success(1) or not */
  char connectAddr[ADDRMAXLN];      /* connected client address */
  char httpStr[BUFFMAXLN];          /* HTTP GET string at terminate */
  int checkInterval;                /* terminal check interval */
  char macAddrNow[ADDRMAXLN];       /* MAC address at now */
  int packetCountPrev=0;            /* packet count at previous check */
  int packetCountNow=0;             /* packet count at now */
  int noPacketPeriod=0;             /* no packet period count in check loop */
  time_t endTime;                   /* the absolute time to exit */
  int leftTime;                     /* the left time length to exit */
  int noPacketInterval=atoi(GetConfValue("ActiveCheck/NoPacketInterval"));  

  checkInterval=atoi(GetConfValue("ActiveCheck/Interval"));

  /* get the time to exit */
  endTime=time(NULL)+duration;

  /* loop until accepting correct user */
  len=sizeof(cliaddr);
  while(1){
    /* alarm setting */
    /* recalculate left time */
    leftTime = endTime - time(NULL);
    if(leftTime <= 0) leftTime=1;
    /* shorter time is set */
    if(leftTime < checkInterval){
      sigfunc=Signal(SIGALRM, timeoutAlarm);
      alarm(leftTime);
    }else{
      sigfunc=Signal(SIGALRM, intervalAlarm);
      alarm(checkInterval);
    }
    if(sigfunc==SIG_ERR){
      connect=0;
      break;
    }
    connfd = -1;

    /* select socket */
    FD_ZERO(&rfd0);
    FD_SET(listenfd[0], &rfd0);
    FD_SET(listenfd[1], &rfd0);

    if(listenfd[0]>listenfd[1]) smax=listenfd[0]+1;
    else smax=listenfd[1]+1;

    //if((n = select(smax, &rfd0, NULL, NULL, &timeout)) > 0){
    if((n = select(smax, &rfd0, NULL, NULL, NULL)) > 0){
      /* wait connection */
      if(FD_ISSET(listenfd[0], &rfd0)){
        if((connfd=accept(listenfd[0], (struct sockaddr *)&cliaddr, &len)) >= 0){
	  /* connect by ipv4 */
          ipType=IPV4;
        }
      }
      if(FD_ISSET(listenfd[1], &rfd0)){
        if((connfd=accept(listenfd[1], (struct sockaddr *)&cliaddr, &len)) >= 0){
	  /* connect by ipv6 */
          ipType=IPV6;
        }
      }
    }
    
    if(connfd < 0){
      /* if usage timeout is reached, end the process */
      if(timeoutAlarmRinged){
        timeoutAlarmRinged=FALSE;
        err_msg("ERR at %s#%d: no connection from applet",__FILE__,__LINE__);
        connect=0;
        break;
      }
      
      /* if check interval is reached, do some checking */
      if(intervalAlarmRinged){
        intervalAlarmRinged=FALSE;
        
        /* packet flow check */
        packetCountNow=GetPacketCount(pClientAddr);
        if(packetCountNow==packetCountPrev){  /* no packet between checks */
          noPacketPeriod++;
        }else{
          noPacketPeriod=0;
          packetCountPrev=packetCountNow;
        }

        if(noPacketPeriod*checkInterval >= noPacketInterval){
          err_msg("ERR at %s#%d: no packet passed for the client",
		  __FILE__,__LINE__);
          connect=0;
          break;
        }
        
        /* mac address check */
        if(ipStatus!=IPV6ONLY){
          GetMacAddrFromArp(clientAddr4, macAddrNow);
          if(strcmp(macAddr4, macAddrNow)!=0){
            err_msg("ERR at %s#%d: mac address is changed",__FILE__,__LINE__);
            connect=0;
            break;
          }
        }
      }
      /* return to connection wait */
      continue;
    }
    
    /* someone is connected */

    /* reset alarm */
    alarm(0);
    Signal(SIGALRM, sigfunc);
    
    /* is it from the correct client addr */
    /* the check is skipped for IPv6 */
    if(ipType==IPV4 && ipStatus!=IPV6ONLY){
      GetPeerAddr(connfd, connectAddr);
      if(isNull(connectAddr)||strcmp(connectAddr, clientAddr4)!=0){
        Close(connfd);
        continue;
      }
    }

    /* set short delay alarm for read in */
    sigfunc=Signal(SIGALRM, delayAlarm);
    if(sigfunc==SIG_ERR){
      connect=0;
      break;
    }

    alarm(COMMWAITTIMEOUT);
    /* get strinxg from connection */
    if(readln(connfd, buff, BUFFMAXLN) <=0){
      if(delayAlarmRinged)delayAlarmRinged=FALSE;
      err_msg("ERR at %s#%d: cannot get userid from applet",__FILE__,__LINE__);
      connect=0;
      break;
    }
    /* some text is read */
    /* reset timeout signal */
    alarm(0);
    Signal(SIGALRM, sigfunc);
    
    /* is it the correct userid from java applet */
    if(strcmp(buff, userid)==0){
      connect=1;
      break;
    }
    
    /* is it the terminate request from accept.html */
    /* the request is [GET /terminate<pid> ....] */
    snprintf(httpStr, BUFFMAXLN, "GET /terminate%d", getpid());
    if(strstr(buff, httpStr)==buff){
      /* terminate request found */
      sendTerminateReply();
      connect=0;
      break;
    }
    
    err_msg("ERR at %s#%d: unknown user or address send from applet",
	    __FILE__,__LINE__);
    
    /* wait new connection */
    Close(connfd);
  }
  
  /* reset timeout signal */
  alarm(0);
  Signal(SIGALRM, sigfunc);
  
  Close(listenfd[0]);
  Close(listenfd[1]);

  return connect;
}

/***************************************/
/* get addr of connected remote site   */
/***************************************/
void getPeerAddr(int sockfd, char *peerAddr)
{
  struct sockaddr *cliaddr;
  socklen_t len;
  char *pAddr;

  *peerAddr='\0'; /* set null string */

  if((cliaddr=Malloc(ADDRMAXLN))==NULL) return; /* if error, return */

  len=ADDRMAXLN;
  if(Getpeername(sockfd, cliaddr, &len)<0) return; /* if error, return */
    
  pAddr=Sock_ntop_host(cliaddr, len);
  if(pAddr!=NULL) strncpy(peerAddr, pAddr, ADDRMAXLN);
  
  free(cliaddr);
  
  return;
}

/***************************************************/
/* wait for close connection                       */
/***************************************************/
void waitClientClose(struct clientAddr *pClientAddr, char *userid, char *userProperty, char *macAddr6, int sipStatus)
{
  char buff[BUFFMAXLN];
  Sigfunc *defaultSigfunc;
  int savePacketCount=0;
  int helloWait=FALSE;
  int noPacketPeriod=0;
  int noReplyCount=0;
  int checkInterval=atoi(GetConfValue("ActiveCheck/Interval"));
  int noPacketInterval=atoi(GetConfValue("ActiveCheck/NoPacketInterval"));  
  int noReplyMaxCount=atoi(GetConfValue("ActiveCheck/NoReplyMaxCount"));

  Writefmt(connfd,"accept\r\n");
  
  /* set interval alarm */
  if((defaultSigfunc=Signal(SIGALRM, intervalAlarm))==SIG_ERR) return;
  alarm(checkInterval);
  
  /* TCP read/write loop */
  /* this loop implement following logic */
  /*   repeat until receiving 'quit' or EOF from client         */
  /*      say 'hello' to client                                 */
  /*      receive 'hello' from client. if no reply, then quit.  */
  /*      get packet count for client after short delay.        */
  /*      long interval                                         */
  /*      get packet count again. if no packet, then quit.      */
  
  while(1){

    if(readln(connfd, buff, BUFFMAXLN)>0){
      /* normal read */

      /* read quit, then quit */
      if(strstr(buff,"quit") != NULL){
        break;
        
        /* read hello, then wait a while for packet log writing */
      }else if((helloWait)&&(strstr(buff,"hello") !=NULL)){
        helloWait=FALSE;
        noReplyCount=0;
        Signal(SIGALRM, delayAlarm);
        alarm(PACKETLOGDELAY);

        /* read any other strings, then do nothing */
      }else{
      }

    }else{
      /* abnormal read */
      
      /* Periodical client check */
      if(intervalAlarmRinged){
        intervalAlarmRinged=FALSE;

        ScanNdpEntry(pClientAddr,userid,macAddr6,userProperty);

        /* if no packet is passed in long time, then quit */
        if(GetPacketCount(pClientAddr)==savePacketCount){
          noPacketPeriod++;

          if(noPacketPeriod*checkInterval >= noPacketInterval){
            err_msg("ERR at %s#%d: no packet passed for the client",
		    __FILE__,__LINE__);
            break;
          }
        }else{
          noPacketPeriod=0;
        }

        /* send hello and wait reply */
        Writefmt(connfd,"hello\r\n");
        helloWait=TRUE;
        Signal(SIGALRM, timeoutAlarm);
        alarm(COMMWAITTIMEOUT);
        
        /* hello reply timeout */
      }else if(timeoutAlarmRinged){
        timeoutAlarmRinged=FALSE;
        
        /* is the reply ignored many times */
        noReplyCount++;
        if(noReplyCount>=noReplyMaxCount){
          /* many times, then quit */
          err_msg("ERR at %s#%d: no reply to hello",__FILE__,__LINE__);
          break;
        }else{
          /* not so many times, then continue */
          helloWait=FALSE;
          savePacketCount=GetPacketCount(pClientAddr);
          Signal(SIGALRM, intervalAlarm);
          alarm(checkInterval);
        }

      /* pass the delay for firewall log renewal */
      }else if(delayAlarmRinged){
        delayAlarmRinged=FALSE;
        savePacketCount=GetPacketCount(pClientAddr);
        Signal(SIGALRM, intervalAlarm);
	alarm(checkInterval);
        
        /* other abnormal read, then quit */
      }else{
        break;
      }
    }
  }
  /* reset alarm */
  Signal(SIGALRM, defaultSigfunc);
  alarm(0);
  SendQuitClient();
  
  return;
}


/***************************************************/
/* send reply to terminate access via connfd(not stdout) */
/***************************************************/
void sendTerminateReply(void)
{
  Writefmt(connfd,"<META HTTP-EQUIV=Pragma CONTENT=no-cache>");
  Writefmt(connfd,"<HTML><BODY> Network is closed. </BODY></HTML> \r\n\r\n");
  Close(connfd);
}

/***************************************************/
/* send quit to client and close connection        */
/***************************************************/
void sendQuitClient(void)
{
  Writefmt(connfd,"quit\r\n");
  Close(connfd);
}

/***************************************************/
/* function called from interval alarm             */
/***************************************************/
static void intervalAlarm(int signo)
{
  intervalAlarmRinged=TRUE;
  return;
}

/***************************************************/
/* function called from timeout alarm              */
/***************************************************/
static void timeoutAlarm(int signo)
{
  timeoutAlarmRinged=TRUE;
  return;
}

/***************************************************/
/* function called from time delay alarm           */
/***************************************************/
static void delayAlarm(int signo)
{
  delayAlarmRinged=TRUE;
  return;
}

/***************************************************/
/***************************************************/
void GetPeerAddr(int sockfd, char *peerAddr)
{
  if(debug) err_msg("DEBUG:=>getPeerAddr( )"); 
  getPeerAddr(sockfd,peerAddr);
  if(debug) err_msg("DEBUG:<=getPeerAddr(,%s)",peerAddr);
}


int GetListenPort(void)
{
  int ret;

  if(debug) err_msg("DEBUG:=>getListenPort( )");
  ret=getListenPort();
  if(debug) err_msg("DEBUG:(%d)<=getListenPort( )",ret);

  return ret;
}

int WaitAppletConnect(char *userid, char *clientAddr4, char *clientAddr6, int duration, char *macAddr4, int ipStatus, struct clientAddr *pClientAddr)
{
  int ret;

  if(debug) err_msg("DEBUG:=>waitAppletConnect(%s,%s,%s,%d,%s,%d)",userid,clientAddr4,clientAddr6,duration,macAddr4,ipStatus);
  ret=waitAppletConnect(userid,clientAddr4,clientAddr6,duration,macAddr4,ipStatus,pClientAddr);
  if(debug) err_msg("DEBUG:(%d)<=waitAppletConnect( )",ret);

  return ret;
}

void WaitClientClose(struct clientAddr *pClientAddr, char *userid, char *userProperty, char *macAddr6, int ipStatus)
{
  if(debug) err_msg("DEBUG:=>waitClientClose(%p,%s,%s,userProperty,%d)",pClientAddr,userid,macAddr6,ipStatus);
  waitClientClose(pClientAddr,userid,userProperty,macAddr6,ipStatus);
  if(debug) err_msg("DEBUG:<=waitClientClose( )");
}

void SendQuitClient(void)
{
  if(debug) err_msg("DEBUG:=>sendQuitClient( )");
  sendQuitClient();
  if(debug) err_msg("DEBUG:<=sendQuitClient( )");
}

void SendTerminateReply(void)
{
  if(debug) err_msg("DEBUG:=>sendTerminateReply( )");
  sendTerminateReply();
  if(debug) err_msg("DEBUG:<=sendTerminateReply( )");
}
