/**************************************************
opengate server main

Copyright (C) 1999 Opengate Project Team
Written by Yoshiaki Watanabe
Modified Katsuhiko Eguchi, 2005 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/

#include	"opengatesrv.h"

extern char ruleNumber4[WORDMAXLN];  /* ipfw rule number in string form  */
extern char ruleNumber6[WORDMAXLN];  /* ip6fw rule number in string form */
extern char language[WORDMAXLN];

char clientAddr4[ADDRMAXLN]="";  /* client addr (nnn.nnn.nnn.nnn) */
char clientAddr6[ADDRMAXLN]="";  /* client addr (nnnn:nnnn:xxxx::xxxx) 128bit */

extern struct clientAddr *pClientAddr;

char macAddr4[ADDRMAXLN]="?";    /* client MAC address (format for arp) */
char macAddr6[ADDRMAXLN]="?";    /* client MAC address (format for ndp) */

char userid[USERMAXLN];
char useridshort[USERMAXLN];/* userID before @ mark(cut off extraID) */
char extraId[USERMAXLN];
char userProperty[BUFFMAXLN];
time_t timeIn, timeOut;

void PutCloseMsg(time_t timeOut, time_t timeIn);
void SetProcessTitle(char *useridshort, char *clientAddr4, char * ruleNumber4, char *clientAddr6, char * ruleNumber6, int ipStatus);
void closeExit(int signo);

/***************************************************/
/*  main routine called as cgi from Web server     */
/***************************************************/
int  main(int argc, char **argv)
{
  char password[PASSMAXLN];
  int port;
  int dummyfd[2];
  int pid;
  int duration;     /* requested usage duration */
  int ipStatus;              /* flag for IPv4 or IPv6 */
  int authResult;

  /* prepare config file */
  if(OpenConfFile()==-1) return 0;
 
  /* start log */
  errToSyslog(atoi(GetConfValue("Syslog/Enable")));
  openlog(GetConfValue("MainCgi"), LOG_PID, atoi(GetConfValue("Syslog/Facility")));

  /* initialize config */
  InitConf();

  /* get default language at the top of lang list */
  sscanf(GetConfValue("HtmlLangs"), "%s", language);

 /* check referer */
  if(CheckReferer()==FALSE){
    PutClientRetry(language);
    return 0;
  }

  /* get POST data */
  if(GetPostData(userid, password, clientAddr4, &duration)==FALSE){
    PutClientRetry(language);
    return 0;
  }

  /* split user@server to user and server */
  SplitId(userid, useridshort, extraId);

  /* setup pointer to ExtraSet in config file */
  SetupConfExtra(useridshort, extraId);

  /* get address of client from getenv. it might be IPv4 or IPv6. */
  GetClientAddr(clientAddr6);

  /* check enable IP versions */
  if((ipStatus=CheckIpVersions(clientAddr4, clientAddr6))==0){
      PutClientMsg("Error: Please contact to the administrator");
      return 0;
  }

  /* get MAC address from arp and ndp */
  GetMacAddr(clientAddr4, macAddr4, clientAddr6, macAddr6,ipStatus);

  /* authenticate the user with authentication server */
  authResult=AuthenticateUser(useridshort, password);
  if(authResult==DENY){
    /* if not authenticate, send deny */
    PutClientDeny(clientAddr4);
    err_msg("DENY: user %s from %s at %s", userid, clientAddr4, macAddr4);
    return 0;
  }

  /* clear password */
  bzero(password, PASSMAXLN);

  /* get user property from user database (if you edit comm-userdb.c) */
  if(!GetUserProperty(userid, userProperty)){
    PutClientMsg("Error: You are denied.");
    err_msg("DENY: user %s from %s at %s (ill-property)",
	    userid, clientAddr4, macAddr4);
    return 0;
  }

  /* usage duration is restricted to permitted range */
  if(duration <= 0){
    duration=atoi(GetConfValue("Duration/Default"));
  }else{
    int durmax=atoi(GetConfValue("Duration/Max"));
    if(duration > durmax){
      duration=durmax;
    }
  }

  /* set terminate signal handler */
  if(Signal(SIGTERM, closeExit)==SIG_ERR){
    PutClientMsg("Error: Please contact to the administrator");    
    return 0;
  }

  /* open firewall for the client */
  if(OpenClientGate(clientAddr4,macAddr4,clientAddr6,macAddr6,
		    userid,userProperty,ipStatus)==FALSE) return 0;
  timeIn=time(NULL);

  /* set (ruleNumber,userid,clientAddr) in process title */
  SetProcessTitle(useridshort,clientAddr4,ruleNumber4,clientAddr6,ruleNumber6,ipStatus);
  
  /* get temporary port for server-listen */
  port=GetListenPort();
  if(port<0){
    err_msg("ERR at %s#%d: cannot get unused listen port",__FILE__,__LINE__);
    PutClientMsg("Error: Please contact to the administrator");
    closeExit(1);
  }

  /* fork */
  if((pid=Fork())==-1){
    err_msg("ERR at %s#%d: fork error",__FILE__,__LINE__);
    PutClientMsg("Error: Please contact to the administrator");
    closeExit(1);
  }

  if(pid!=0){
    /** parent process **/
    /* send java */
    PutJavaApplet(userid, port, pid, clientAddr4, clientAddr6, ipStatus);

    /* detach from Web server */
    return 0;
  }
  /** child process **/

  /* detach from Web server */
  Close(0);Close(1);  /* detach stdin and out pipe connected to Web */
  Close(2);           /* detach stderr */
  Pipe(dummyfd);      /* connect dummy pipe for stdin and out */

  /* wait connection from the java applet */
  /* if no connection, close gate when duration is passed */
  /* or ipaddr for the macAddr4 is changed */
  if(WaitAppletConnect(userid, clientAddr4, clientAddr6, duration, macAddr4, ipStatus, pClientAddr)==1){

    /* wait until the user quit */
    /* close gate when no reply to temporal hello or end of TCP connection */
    /* macAddr6 is used for NDP to search addition ipaddr */
    WaitClientClose(pClientAddr, userid, userProperty, macAddr6, ipStatus);
  }
  /* close firewall and exit */
  closeExit(1);

  return 0;
}


/*******************************************/
/* calc connection duration and put it out */
/*******************************************/
void putCloseMsg(time_t timeOut, time_t timeIn)
{
  double time;
  int hour,min,sec;

  time=difftime(timeOut,timeIn);
  hour=time/60/60;
  min=(time-hour*60*60)/60;
  sec=(time-hour*60*60-min*60);
  //err_msg("STOP: user %s at %s ( %02d:%02d:%02d )", userid, macAddr4, hour,min,sec);
  return;
}


/*****************************/
/* At termination, call this */
/*****************************/
void closeExit(int signo)
{
  /* signal is disabled */
  Signal(SIGTERM, SIG_DFL);

  /* send quit message to client java */
  SendQuitClient();

  /* close firewalls */
  while(pClientAddr!=NULL){
    if(pClientAddr->ipType==IPV4){
      CloseClientGate4(pClientAddr,userid,macAddr4);
    }else{
      CloseClientGate6(pClientAddr,userid,macAddr6);
      DeleteNdpEntry(pClientAddr->ipAddr);
    }
    pClientAddr = pClientAddr->next;
  }

  /* put out time */
  timeOut=time(NULL);
  PutCloseMsg(timeOut,timeIn);
  if(debug) err_msg("DEBUG:terminated");

  exit(1);
}

/************************************/
/* set up process title string      */
/************************************/
void setProcessTitle(char *useridshort, char *clientAddr4, char * ruleNumber4, char *clientAddr6, char * ruleNumber6, int ipStatus)
{
  switch(ipStatus){
  case IPV46DUAL:
    setproctitle("%s(useIPv6),[%s(%s)],[%s(%s)]", useridshort, 
		 clientAddr4, ruleNumber4, clientAddr6, ruleNumber6);
    break;

  case IPV4ONLY:
    setproctitle("%s,[%s(%s)],",useridshort, clientAddr4, ruleNumber4);
    break;

  case IPV6ONLY:
    setproctitle("%s(useIPv6),,[%s(%s)]",useridshort,  clientAddr6, ruleNumber6);
    break;

  default:
    err_msg("ERR at %s#%d: abnormal IP versions %d",__FILE__,__LINE__,ipStatus);
  }
}
/*****************************/
/*****************************/
void PutCloseMsg(time_t timeOut, time_t timeIn)
{
  if(debug) err_msg("DEBUG:=>putCloseMsg( )");
  putCloseMsg(timeOut,timeIn);
  if(debug) err_msg("DEBUG:<=putCloseMsg( )");
}

void SetProcessTitle(char *useridshort, char *clientAddr4, char * ruleNumber4, char *clientAddr6, char * ruleNumber6, int ipStatus){
  if(debug) err_msg("DEBUG:=>setProcessTitle(%s,%s,%s,%s,%s,%d)",useridshort,
		    clientAddr4,ruleNumber4,clientAddr6,ruleNumber6, ipStatus);
  setProcessTitle(useridshort,clientAddr4,ruleNumber4,clientAddr6,ruleNumber6,ipStatus);
  if(debug) err_msg("DEBUG:<=setProcessTitle( )");
}
